from __future__ import annotations

import re

from antenna_lib.array_antennas.array import AntennaArray
from antenna_lib.array_antennas.broadside import BroadSideArray
from antenna_lib.array_antennas.hansen_woodyard import HansenWoodyardArray
from antenna_lib.array_antennas.nonuniform_af import NonUniformAntennaArray
from antenna_lib.array_antennas.ordinary_endfire import OrdinaryEndFireArray
from antenna_lib.array_antennas.uniform_af import UniformAntennaArray
from antenna_lib.single_antennas import SingleAntenna


class AntennaArrayFactory:
    @classmethod
    def create_array(cls,
                     antenna_elem: SingleAntenna,
                     spacing: float = 0.5,
                     array_type: str | None = None,
                     phase_progression: float = 0.0,
                     n_elements: None | int = None,
                     amplitudes: None | list[float] = None) -> AntennaArray:
        if amplitudes:
            return NonUniformAntennaArray(antenna_elem, amplitudes, spacing, phase_progression)
        elif n_elements and not amplitudes:
            if array_type:
                if match := re.match(r'endfire-(\d{1,3})', array_type):
                    return OrdinaryEndFireArray(antenna_elem, n_elements, spacing, float(match.group(1)))
                elif 'broadside' == array_type:
                    return BroadSideArray(antenna_elem, n_elements, spacing)
                elif match := re.match(r'hansen-woodyard-(0|180)$', array_type):
                    return HansenWoodyardArray(antenna_elem, n_elements, spacing, match.group(1))
                else:
                    raise ValueError("the array type can be `endfire-{angle}`, `broadside`,"
                                     " or `hansen-woodyard-{0-180}`")
            return UniformAntennaArray(antenna_elem, n_elements, spacing, phase_progression)
        else:
            raise ValueError("Need to specify either `amplitudes: list` or `n_elements: int`")



from functools import wraps

import numpy as np


def rotatory(r):
    """
    Decorador que permite rotar cualquier patron de radiación sobre el eje x.
    :param r: directivity method
    :return: rotated function
    """
    @wraps(r)
    def wrapper(self, theta: float, phi: float = 0.0):
        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)
        x_prime = np.round(x, 10)
        y_prime = np.round(y * np.cos(self.angle) + z * np.sin(self.angle), 10)
        z_prime = np.round(z * np.cos(self.angle) - y * np.sin(self.angle), 10)
        theta_prime = np.arccos(z_prime)
        phi_prime = np.arctan2(y_prime, x_prime)
        return r(self, theta_prime, phi_prime)
    return wrapper



import numpy as np

from antenna_lib.antenna import Antenna
from antenna_lib.array_antennas.array import AntennaArray
from antenna_lib.array_antennas.uniform_af import UniformAntennaArray
from antenna_lib.single_antennas import DipoleAntenna, LoopAntenna, SingleAntenna


def test_create_single_antenna():
    dipole = DipoleAntenna(length=0.5, pol=0.0)
    assert isinstance(dipole, DipoleAntenna)
    assert isinstance(dipole, SingleAntenna)
    assert isinstance(dipole, Antenna)
    assert dipole.length == 0.5
    assert hasattr(dipole, 'polarization')


def test_create_array_antenna():
    loop = LoopAntenna(radius=1.2)
    arr_antenna = UniformAntennaArray(loop, n_elements=3)
    assert isinstance(arr_antenna, AntennaArray)
    assert isinstance(arr_antenna, Antenna)
    assert arr_antenna.n_elements == 3
    assert hasattr(arr_antenna, 'polarization')
    assert np.allclose(arr_antenna.polarization.pol_vector, np.array([0, 1]), 1e-6)



from __future__ import annotations

import warnings

import numpy as np

from antenna_lib.antenna_parameters import PolarizationFactory
from antenna_lib.single_antennas.single import SingleAntenna
from antenna_lib.utils.decorators import rotatory


class DipoleAntenna(SingleAntenna):

    def __init__(self,
                 length: float, pol: str | float = 0.0,
                 amplitude: float = 1.0):
        if length <= 0:
            raise ValueError('Dipole length must be greater than zero.')
        self.length = length
        if isinstance(pol, str):
            if pol == 'horizontal':
                pol = 90.0
            elif pol == 'vertical':
                pol = 0.0
            else:
                raise ValueError('Invalid polarization string')
        self.angle = pol * np.pi / 180
        self.polarization = PolarizationFactory.create_polarization(f'linear@{pol}')
        super().__init__(amplitude=amplitude)

    @property
    def max_directivity(self) -> float:
        """Calculate max directivity"""
        if self.length <= 0.1:
            return 1.5
        elif self.length >= 1.0:
            return super().max_directivity
        else:
            return self.directivity(np.pi / 2 - self.angle, 0.0)

    @rotatory
    def field_pattern(self, theta: float, phi: float = 0.0) -> float:
        kl = 2 * np.pi * self.length
        if self.length <= 0.1:
            return np.sin(theta)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            e = ((np.cos(kl / 2 * np.cos(theta)) - np.cos(kl / 2)) / np.sin(theta))
        return 0.0 if np.isnan(e) else np.abs(e)



import numpy as np

from antenna_lib.exceptions import InvalidPolarizationException


class Polarization:
    def __init__(self,
                 pol_vector: np.ndarray,
                 polarization_str: str = None):
        """
        Inicializa la clase Polarization, con un vector de polarización (unitario) y una representación
        en string de su polarización.
        :param pol_vector: Vector complejo unitario que representa la polarización de la onda transmitida de la antena.
        :param polarization_str: Representación en string.
        """
        self.pol_vector = pol_vector
        self._polarization_str = polarization_str

    @property
    def polarization_str(self):
        """Lógica para detectar el tipo de polarización basado en el vector de polarización.
        Necesario para el uso con arreglos de antenas."""
        if self._polarization_str is None:
            # Remember pol_vector has the form: x * E_x * exp(j * phi_x) + y * E_y * exp(j * phi_y)
            y_over_x = self.pol_vector[1] / self.pol_vector[0]
            if np.isreal(y_over_x) or np.isclose(self.pol_vector[0], 0.0 + 0.0j):
                # it's linear
                y_over_x = np.real(y_over_x)  # for the case of getting 0j
                return f'linear@{np.arctan(y_over_x) * 180 / np.pi}'
            elif np.isclose(y_over_x, 1j) or np.isclose(y_over_x, -1j):
                # it's circular
                return f'circular@{"rcp" if np.isclose(y_over_x, 1j) else "lcp"}'
            else:
                # it's elliptical
                gamma = np.arctan(np.abs(y_over_x))
                delta = np.angle(y_over_x)
                epsilon = np.arcsin(np.sin(2 * gamma) * np.sin(delta)) / 2
                ar = (1 if delta > 0 else -1) / np.tan(epsilon)
                tau = np.arctan(np.tan(2 * gamma) * np.cos(delta)) / 2
                return f'elliptical@{"rcp" if delta > 0 else "lcp"}@{round(ar, 2)}@{round(tau * 180 / np.pi, 2)}'
        return self._polarization_str

    def __rmul__(self, other):
        return other * self.pol_vector

    def __repr__(self):
        return f'<{self.polarization_str}, Polarization vector: {self.pol_vector.round(2)}>'


class PolarizationFactory:
    """Clase encargada de crear objetos de tipo Polarization"""

    @classmethod
    def create_polarization(cls, polarization_str: str) -> Polarization:
        pol_vector = cls.parse_pol_vector(polarization_str)
        return Polarization(pol_vector, polarization_str)

    @classmethod
    def parse_pol_vector(cls, polarization_str: str) -> np.ndarray:
        if polarization_str.startswith('linear'):
            try:
                polarization_angle = float(polarization_str.split('@')[1]) * np.pi / 180
            except IndexError:
                raise IndexError('Invalid parameter for linear polarization angle. Format is "linear@{angle}"')
            except ValueError:
                raise ValueError('Please specify a valid float number as an angle')
            return np.array([np.cos(polarization_angle), np.sin(polarization_angle)])
        elif polarization_str.startswith('circular'):
            try:
                polarization_orientation = polarization_str.split('@')[1]
                if polarization_orientation not in ['lcp', 'rcp']:
                    raise InvalidPolarizationException('No valid polarization orientation. Valid parameters are '
                                                       '"rcp", "lcp"')
                return np.array([1, 1j if polarization_orientation == 'rcp' else -1j]) / np.sqrt(2)
            except IndexError:
                raise IndexError('Invalid parameter for circular polarization orientation. Format is '
                                 'circular@{orientation}')
        elif polarization_str.startswith('elliptical'):
            try:
                orientation = polarization_str.split('@')[1]
                ar = float(polarization_str.split('@')[2])
                tau = float(polarization_str.split('@')[3]) * np.pi / 180

                if orientation not in ['rcp', 'lcp']:
                    raise InvalidPolarizationException('No valid Polarization orientation. Valid parameters are '
                                                       '"rcp", "lcp"')
                # calculate delta
                epsilon = np.arctan(1 / ar if orientation == 'rcp' else 1 / -ar)
                delta = np.arctan2(np.tan(2 * epsilon), np.sin(2 * tau))
                # calculate E_y / E_x
                cos_2gamma = np.cos(2 * epsilon) * np.cos(2 * tau)
                gamma = np.arccos(cos_2gamma) / 2
                ey_ex = np.tan(gamma)
                pol = np.array([1, ey_ex * np.exp(1j * delta)])
                return pol / np.linalg.norm(pol)
            except IndexError:
                raise IndexError('Invalid parameter for elliptical polarization. Format is '
                                 'elliptical@{orientation}@{axial_ratio}@{tau}')
            except ValueError:
                raise ValueError('Please specify a valid float number')
        else:
            raise InvalidPolarizationException("No valid polarization")



import numpy as np

from antenna_lib.antenna_parameters import Polarization
from antenna_lib.array_antennas.array import AntennaArray
from antenna_lib.single_antennas import SingleAntenna


class NonUniformAntennaArray(AntennaArray):

    def __init__(self, antenna: SingleAntenna, amplitudes: list[float], spacing: float = 0.5,
                 phase_progression: float = 0.0):
        super().__init__(antenna, spacing, phase_progression)
        self.amplitudes = amplitudes
        self.n_elements = len(amplitudes)

    @property
    def polarization(self) -> Polarization:
        if self._polarization is None:
            pol = sum(
                [(amplitude * self.antenna.polarization) * np.exp(1j * k * self.phase_progression)
                 for k, amplitude in enumerate(self.amplitudes)])
            self._polarization = Polarization(pol / np.linalg.norm(pol))
            return self._polarization
        return self._polarization

    def array_factor(self, theta: float, phi: float) -> float:
        kd = 2 * np.pi * self.spacing
        psi = kd * np.cos(theta) + self.phase_progression
        return np.abs(sum(amplitude * np.exp(1j * n * psi) for n, amplitude in enumerate(self.amplitudes)))



from abc import abstractmethod

from antenna_lib.antenna import Antenna
from antenna_lib.antenna_parameters import Polarization
from antenna_lib.single_antennas import SingleAntenna


class AntennaArray(Antenna):
    def __init__(self,
                 antenna: SingleAntenna,
                 spacing: float = 0.5,
                 phase_progression: float = 0.0):
        super().__init__()
        self.antenna = antenna
        self.spacing = spacing
        self.phase_progression = phase_progression
        self.n_elements = None
        self.amplitudes = None
        self._polarization = None

    def field_pattern(self, theta, phi):
        return self.antenna.field_pattern(theta, phi) * self.array_factor(theta, phi)

    @abstractmethod
    def array_factor(self, theta: float, phi: float) -> float:
        pass

    @property
    @abstractmethod
    def polarization(self) -> Polarization:
        """Calcular la polarización resultante a partir de las antenas que conforman el array"""
        pass

    def play_wave_animation(self):
        """Implementation for this antenna type"""



import numpy as np

from antenna_lib.antenna_parameters import Polarization
from antenna_lib.array_antennas.array import AntennaArray
from antenna_lib.single_antennas import SingleAntenna


class UniformAntennaArray(AntennaArray):
    def __init__(self, antenna: SingleAntenna, n_elements: int, spacing: float = 0.5,
                 phase_progression: float = 0.0):
        super().__init__(antenna, spacing, phase_progression)
        self.n_elements = n_elements

    @property
    def polarization(self) -> Polarization:
        if self._polarization is None:
            pol = sum(
                [np.exp(1j * k * self.phase_progression) * self.antenna.polarization for k in range(self.n_elements)])
            self._polarization = Polarization(pol / np.linalg.norm(pol))
            return self._polarization
        return self._polarization

    def array_factor(self, theta: float, phi: float) -> float:
        kd = 2 * np.pi * self.spacing
        psi = kd * np.cos(theta) + self.phase_progression
        return np.abs(sum(np.exp(1j * n * psi) for n in range(self.n_elements)))



import numpy as np

from antenna_lib.array_antennas.uniform_af import UniformAntennaArray
from antenna_lib.single_antennas import SingleAntenna


class OrdinaryEndFireArray(UniformAntennaArray):

    def __init__(self, antenna: SingleAntenna, n_elements: int, spacing: float = 0.5, angle: float = 0.0):
        super().__init__(antenna, n_elements, spacing)
        self.phase_progression = - 2 * np.pi * spacing * np.cos(angle * np.pi / 180)



import numpy as np

from antenna_lib.single_antennas.single import SingleAntenna
from scipy.special import jv


class LoopAntenna(SingleAntenna):

    def __init__(self,
                 radius: float,
                 polarization: str = 'linear@90.0',
                 amplitude: float = 1.0):
        self.radius = radius
        super().__init__(polarization, amplitude)

    def field_pattern(self, theta: float, phi: float = 0.0) -> float:
        if self.radius <= 1 / (6 * np.pi):
            return np.sin(theta)
        return jv(1, 2 * np.pi * self.radius * np.sin(theta))



import warnings
from abc import abstractmethod, ABC
from functools import partial

import numpy as np
from matplotlib import pyplot as plt
from scipy.integrate import dblquad


class Antenna(ABC):
    polarization = None
    amplitude = None
    _radiated_power = None

    @abstractmethod
    def field_pattern(self, theta, phi):
        """Patrón de campo de la antena"""

    def power_pattern(self, theta: float, phi: float) -> float:
        """Patrón de potencia, cuadrado del patrón de campo"""
        return self.field_pattern(theta, phi) ** 2

    @property
    def radiated_power(self) -> float:
        """Potencia radiada en todas las direcciones"""
        if self._radiated_power is None:
            f = lambda t, p: (self.field_pattern(t, p) ** 2) * np.sin(t)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._radiated_power = dblquad(f, 0, 2 * np.pi, 0, np.pi)[0]
        return self._radiated_power

    def directivity(self, theta: float, phi: float) -> float:
        """Función directividad calculada a partir del patrón de potencia y la potencia radiada"""
        return 4 * np.pi * self.power_pattern(theta, phi) / self.radiated_power

    @property
    def max_directivity(self) -> float:
        """Implementación por defecto de la directividad máxima"""
        theta = np.linspace(0, np.pi, 100)
        phi = np.linspace(0, 2 * np.pi, 100)
        theta, phi = np.meshgrid(theta, phi)
        r = np.vectorize(self.directivity)(theta, phi)
        return np.max(r)

    def _horizontal_vertical_patterns(self, theta: np.ndarray, phi: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        horizontal = np.array(list(map(partial(self.power_pattern, np.pi / 2), phi)))
        vertical = np.array(list(map(partial(self.power_pattern, phi=np.pi / 2), theta)) +
                            list(map(partial(self.power_pattern, phi=3 * np.pi / 2), np.flip(theta))))
        return horizontal, vertical

    def plot_radiation_pattern(self, polar=True, field=False, log_scale=False):
        phi = np.linspace(0, 2 * np.pi, 200)
        theta_pattern = np.linspace(0, np.pi, 100)
        horizontal, vertical = self._horizontal_vertical_patterns(theta_pattern, phi)
        theta = np.linspace(0, 2 * np.pi, 200)
        if field:
            horizontal, vertical = np.sqrt(horizontal), np.sqrt(vertical)
        horizontal /= np.max(horizontal)
        vertical /= np.max(vertical)
        plt.subplots_adjust(wspace=0.4)
        h_plane = plt.subplot(1, 2, 1, projection='polar')
        v_plane = plt.subplot(1, 2, 2, projection='polar')
        if log_scale:
            horizontal = 10 * np.log10(1e-4 + horizontal)
            vertical = 10 * np.log10(1e-4 + vertical)
            h_plane.set_rlim(-40)
            v_plane.set_rlim(-40)
        h_plane.set_theta_zero_location('N')
        h_plane.set_theta_direction(-1)
        h_plane.plot(phi, horizontal)
        v_plane.set_theta_zero_location('N')
        v_plane.set_theta_direction(-1)
        v_plane.plot(theta, vertical)
        plt.show()

    def plot_3d_pattern(self, field=False):
        """Gráfico 3d del patrón de radiación."""
        theta, phi = np.meshgrid(np.linspace(0, np.pi, 100), np.linspace(0, 2 * np.pi, 100))
        r = np.vectorize(self.field_pattern)(theta, phi) ** (1 if field else 2)
        x = r * np.sin(theta) * np.cos(phi)
        y = r * np.sin(theta) * np.sin(phi)
        z = r * np.cos(theta)
        fig, ax = plt.subplots(subplot_kw={'projection': '3d'})
        # to have same length in all axes
        max_range = np.array([x.max() - x.min(), y.max() - y.min(), z.max() - z.min()]).max() / 2.0
        mid_x = (x.max() + x.min()) * 0.5
        mid_y = (y.max() + y.min()) * 0.5
        mid_z = (z.max() + z.min()) * 0.5
        ax.set_xlim(mid_x - max_range, mid_x + max_range)
        ax.set_ylim(mid_y - max_range, mid_y + max_range)
        ax.set_zlim(mid_z - max_range, mid_z + max_range)
        # cmap according to r value
        norm = plt.Normalize(0, np.max(r))
        colors = norm(r)
        cmap = plt.cm.get_cmap('jet')
        ax.plot_surface(x, y, z, alpha=0.8, rstride=1, cstride=1, cmap='jet', facecolors=cmap(colors),
                        linewidth=0, shade=True)
        plt.show()



import numpy as np

from antenna_lib.single_antennas import DipoleAntenna
import pytest


def test_dipole_init():
    dip = DipoleAntenna(length=0.5)
    assert hasattr(dip, 'polarization')
    assert hasattr(dip, 'amplitude')
    assert pytest.approx(dip.length) == 0.5
    with pytest.raises(ValueError) as error_info:
        dip2 = DipoleAntenna(length=-0.5)
    assert str(error_info.value) == 'Dipole length must be greater than zero.'


@pytest.mark.parametrize('pol, angle', [('horizontal', 90.0), ('vertical', 0.0), ('linear', None), (45.0, 45.0)])
def test_dipole_multiple_init_pols(pol, angle):
    if pol == 'linear':
        with pytest.raises(ValueError):
            dip = DipoleAntenna(length=0.5, pol=pol)
    else:
        dip = DipoleAntenna(length=0.5, pol=pol)
        assert pytest.approx(dip.angle) == angle * np.pi / 180


def test_dipole_repr():
    dip = DipoleAntenna(length=0.5)
    assert str(dip).startswith('<DipoleAntenna with polarization:')


@pytest.mark.parametrize('dip_length, max_directivity', [(0.05, 1.5), (1.5, 2.226), (0.5, 1.641)])
def test_dipole_max_directivity(dip_length, max_directivity):
    dip = DipoleAntenna(length=dip_length)
    assert pytest.approx(dip.max_directivity, 1e-2) == max_directivity
    with pytest.raises(AttributeError):
        # noinspection PyPropertyAccess
        dip.max_directivity = 1.0


def test_dipole_field_pattern():
    shortwave_dipole = DipoleAntenna(0.01)
    for theta in np.linspace(0, np.pi, 10):
        assert pytest.approx(shortwave_dipole.field_pattern(theta), 1e-4) == np.sin(theta)
    halfwave_dipole = DipoleAntenna(0.5)
    assert pytest.approx(halfwave_dipole.field_pattern(0.0)) == 0.0
    assert pytest.approx(halfwave_dipole.field_pattern(np.pi / 2)) == 1.0



from .dipole import DipoleAntenna
from .loop import LoopAntenna
from .single import SingleAntenna
from .isotropic import IsotropicAntenna

__all__ = [DipoleAntenna, LoopAntenna, SingleAntenna, IsotropicAntenna]



import numpy as np

from antenna_lib.array_antennas.uniform_af import UniformAntennaArray
from antenna_lib.single_antennas import SingleAntenna


class HansenWoodyardArray(UniformAntennaArray):

    def __init__(self, antenna: SingleAntenna, n_elements: int, spacing: float = 0.5, up: str = '0'):
        super().__init__(antenna, n_elements, spacing)
        self.phase_progression = (2 * np.pi * spacing + 2.92 / n_elements) * (-1 if up == '0' else 1)



from antenna_lib.antenna_parameters import PolarizationFactory
from antenna_lib.single_antennas.single import SingleAntenna


class IsotropicAntenna(SingleAntenna):

    """Antena isotropica. Puede tener cualquier polarización. Su patrón es siempre omnidireccional"""

    def __init__(self,
                 polarization: str = 'linear@0.0',
                 amplitude: float = 1.0):
        self.polarization = PolarizationFactory.create_polarization(polarization)
        super().__init__(amplitude)

    @property
    def max_directivity(self) -> float:
        return 1.0

    def field_pattern(self, theta: float, phi: float = 0.0) -> float:
        """Patrón de campo de antena insotropica."""
        return 1.0



import pytest

from antenna_lib.array_antennas.factory import AntennaArrayFactory
from antenna_lib.single_antennas import IsotropicAntenna


def test_array_factory():
    # define antenna
    antenna = IsotropicAntenna()
    antenna_array = AntennaArrayFactory.create_array(antenna_elem=antenna,
                                                     spacing=3,
                                                     phase_progression=0.0,
                                                     amplitudes=[0.5, 1.5])
    assert antenna_array.n_elements == 2
    assert pytest.approx(antenna_array.spacing) == 3
    assert hasattr(antenna_array, 'phase_progression')


def test_polarization_phase_progression():
    pass



import numpy as np
import pytest

from antenna_lib.antenna_parameters import PolarizationFactory
from antenna_lib.array_antennas import AntennaArrayFactory
from antenna_lib.exceptions import InvalidPolarizationException
from antenna_lib.single_antennas.isotropic import IsotropicAntenna


def test_polarization_factory():
    linear_pol = PolarizationFactory.create_polarization('linear@90')
    assert np.allclose(linear_pol.pol_vector, np.array([0, 1]), 1e-6)
    assert str(linear_pol) == '<linear@90, Polarization vector: [0. 1.]>'

    circular_pol = PolarizationFactory.create_polarization('circular@lcp')
    assert np.allclose(circular_pol.pol_vector, np.array([1, -1j]) / np.sqrt(2), 1e-6)
    assert str(circular_pol) == '<circular@lcp, Polarization vector: [ 0.71+0.j   -0.  -0.71j]>'

    ellip_pol = PolarizationFactory.create_polarization('elliptical@rcp@1.5@45')
    assert np.allclose(ellip_pol.pol_vector, np.array([0.70710678, 0.27196415 + 0.65271395j]), 1e-6)
    assert str(ellip_pol) == '<elliptical@rcp@1.5@45, Polarization vector: [0.71+0.j   0.27+0.65j]>'


def test_invalid_polarizations_linear():
    with pytest.raises(InvalidPolarizationException) as error_info:
        invalid_pol = PolarizationFactory.create_polarization('poly')
    assert str(error_info.value) == 'No valid polarization'

    with pytest.raises(IndexError) as error_info:
        invalid_format = PolarizationFactory.create_polarization('linear')
    assert str(error_info.value) == 'Invalid parameter for linear polarization angle. Format is "linear@{angle}"'

    with pytest.raises(ValueError) as error_info:
        invalid_angle = PolarizationFactory.create_polarization('linear@pi/2')
    assert str(error_info.value) == 'Please specify a valid float number as an angle'


def test_invalid_polarizations_circular():
    with pytest.raises(IndexError) as error_info:
        invalid_format = PolarizationFactory.create_polarization('circular')
    assert str(error_info.value) == ('Invalid parameter for circular polarization orientation. Format is '
                                     'circular@{orientation}')

    with pytest.raises(InvalidPolarizationException) as error_info:
        invalid_rotation = PolarizationFactory.create_polarization('circular@50')
    assert str(error_info.value) == 'No valid polarization orientation. Valid parameters are "rcp", "lcp"'


def test_invalid_polarizations_elliptical():
    with pytest.raises(InvalidPolarizationException) as error_info:
        invalid_rotation = PolarizationFactory.create_polarization('elliptical@lhcp@3@45')
    assert str(error_info.value) == 'No valid Polarization orientation. Valid parameters are "rcp", "lcp"'

    with pytest.raises(IndexError) as error_info:
        invalid_format = PolarizationFactory.create_polarization('elliptical@rcp@50')
    assert str(error_info.value) == ('Invalid parameter for elliptical polarization. Format is '
                                     'elliptical@{orientation}@{axial_ratio}@{tau}')

    with pytest.raises(ValueError) as error_info:
        invalid_values = PolarizationFactory.create_polarization('elliptical@rcp@p10@pi')
    assert str(error_info.value) == 'Please specify a valid float number'


@pytest.mark.parametrize('pol, pol_vector', [('linear@0.0', [1, 0]), ('circular@rcp', [0.7071068, 0.7071068j]),
                                             ('elliptical@rcp@1.5@45.0', [0.70710678, 0.27196415 + 0.65271395j])])
def test_array_polarization_str(pol, pol_vector):
    antenna = IsotropicAntenna(polarization=pol)
    arr_antenna = AntennaArrayFactory.create_array(antenna, n_elements=3)
    assert np.allclose(arr_antenna.polarization.pol_vector, np.array(pol_vector), 1e-6)
    assert arr_antenna.polarization.polarization_str == pol



from unittest.mock import patch, MagicMock, call

import numpy as np
import pytest

from antenna_lib.exceptions import FieldPatternNotImplementedException
from antenna_lib.single_antennas import SingleAntenna


class MyTestAntenna(SingleAntenna):
    """A super special antenna with a defined field pattern for testing"""
    def field_pattern(self, theta: float, phi: float = 0.0) -> float:
        if theta <= np.pi / 2:
            return np.cos(theta) ** 2
        return 0


class MyInvalidAntenna(SingleAntenna):
    """Doesn't implement field pattern method"""
    pass


def test_single_antenna_init():
    antenna = MyTestAntenna()
    assert antenna.polarization.polarization_str == 'linear@0.0'
    assert pytest.approx(antenna.amplitude) == 1.0
    unimp_antenna = MyInvalidAntenna()
    with pytest.raises(FieldPatternNotImplementedException):
        unimp_antenna.field_pattern(0.0)


def test_single_antenna_directivity():
    antenna = MyTestAntenna()
    assert pytest.approx(antenna.max_directivity) == 10.0
    assert pytest.approx(antenna.directivity(0.0, 0.0)) == 10.0
    assert pytest.approx(antenna.directivity(np.pi / 2, 0.0)) == 0.0
    assert pytest.approx(antenna.directivity(np.pi, 0.0)) == 0.0


@patch('antenna_lib.antenna.plt')
@patch('antenna_lib.antenna.np')
def test_single_antenna_plot_radiation_pattern(mock_numpy: MagicMock, mock_pyplot: MagicMock):
    antenna = MyTestAntenna()
    antenna.plot_radiation_pattern()
    mock_pyplot.title.assert_not_called()
    mock_pyplot.subplot.assert_has_calls([call(1, 2, 1, projection='polar'), call(1, 2, 2, projection='polar')])
    mock_pyplot.subplot().set_theta_zero_location.assert_has_calls([call('N'), call('N')])
    mock_pyplot.subplot().set_theta_direction.assert_has_calls([call(-1), call(-1)])
    mock_pyplot.show.assert_called_once()
    # test with field
    antenna.plot_radiation_pattern(field=True)
    assert len(mock_numpy.sqrt.call_args_list) == 2

    # test with log_scale
    antenna.plot_radiation_pattern(log_scale=True)
    assert len(mock_numpy.log10.call_args_list) == 2
    mock_pyplot.subplot().set_rlim.assert_has_calls([call(-40), call(-40)])


@patch('antenna_lib.antenna.plt')
@patch('antenna_lib.antenna.np')
def test_single_antenna_plot_3d_pattern(mock_numpy: MagicMock, mock_pyplot: MagicMock):
    mock_numpy.meshgrid.return_value = MagicMock(), MagicMock()
    mock_pyplot.subplots.return_value = MagicMock(), MagicMock()
    antenna = MyTestAntenna()
    antenna.plot_3d_pattern()
    mock_pyplot.subplots.assert_called_with(subplot_kw={'projection': '3d'})
    mock_pyplot.subplots()[1].plot_surface.assert_called_once()
    mock_pyplot.show.assert_called_once()



from __future__ import annotations

from antenna_lib.antenna import Antenna
from antenna_lib.antenna_parameters import PolarizationFactory
from antenna_lib.exceptions import FieldPatternNotImplementedException


class SingleAntenna(Antenna):

    def __init__(self,
                 pol: str | float = 'linear@0.0',
                 amplitude: float = 1.0):
        super().__init__()
        self.amplitude = amplitude
        if self.polarization is None:
            self.polarization = PolarizationFactory.create_polarization(pol)
        self._radiated_power = None

    def field_pattern(self, theta: float, phi: float = 0.0) -> float:
        """Patrón de campo. Propiedad a partir de la cual se obtienen muchas de las características de la antena"""
        raise FieldPatternNotImplementedException('The antenna does not have a `field_pattern`'
                                                  ' method implemented')

    def __repr__(self):
        return f'<{self.__class__.__name__} with polarization:\n{self.polarization}'
